/*
 
		A short program that verifies the existence of period orbits for 
		the Lorenz equations. The validation uses the backward shooting 
		method of Galias, but solves the sparse linear system via Gauss 
		elimination. It reads multiprecision initial data from the input 
		file "/data/lor_N.txt", where N is the period given as an argument.
 

		Compilation: make wtrb
		Syntax:  ./wtrb [radius] [bits] [period] [filename]
		Example: ./wtrb 1e-80 400 2 lor_2_BA.txt

		Author: Warwick Tucker <warwick@math.uu.se>
		Latest edit: Mon Feb 20 12:48:17 CET 2012
    
*/

#include <iostream>
#include <fstream>
#include <stdio.h>

#include "capd/mpcapdlib.h"
#include "capd/dynset/C0Rect2RSet.hpp"
#include "capd/poincare/PoincareMapJet.h"

#define ORDER 90 // Order of the Taylor ODE integrator.

#undef PRINTOUT
//#define PRINTOUT

enum command {START_TIMING, SHOW_TIMING, STOP_TIMING};

using namespace capd;
using namespace std;

// Multi-precision typedefs --------------------------------------------
typedef capd::multiPrec::MpReal real;
typedef capd::vectalg::Vector<real, 0>   RVector;    // MP vector
typedef capd::vectalg::Matrix<real, 0,0> RMatrix;    // MP matrix
typedef MpInterval interval;
typedef capd::MpIVector IVector;
typedef capd::MpIMatrix IMatrix;
typedef capd::MpIFunction IFunction;
typedef capd::MpIMap IMap;
typedef capd::MpITaylor ITaylor;
typedef capd::poincare::PoincareMapJet<capd::MpITaylor> PoincareMap;
typedef capd::dynset::C0Rect2RSet<IMatrix> C0Rect2Set;
typedef capd::MpC1Rect2 C1Rect2Set;
typedef capd::MpIEuclLNorm IEuclLNorm;

// A basic timer--------------------------------------------------------
static void clock(const command &cmd)
{
  struct tm *ptr;              // Used for timing the program 
  static time_t time_start;    // start time 

  if ( cmd == START_TIMING ) {
		time_start = time(NULL);
     ptr = localtime(&time_start); 
     cout << "Started computations: " << asctime(ptr);  
  }
  else if ( cmd == SHOW_TIMING || cmd == STOP_TIMING ) {
    int weeks   = 0;
    int days    = 0;
    int hours   = 0;
    int minutes = 0;
    double seconds;
    time_t time_stop;  

    time_stop = time(NULL);
    ptr = localtime(&time_stop);
    if ( cmd == STOP_TIMING )
			cout << "Completed computations: " << asctime(ptr);
    seconds = difftime(time_stop, time_start);
    cout << "Elapsed run time: "; 
    // Convert seconds to vv:dd:hh:mm:ss
    while ( seconds >= 60 ) {
			seconds -= 60;
			minutes++;
		}
    while ( minutes >= 60 ) {
			minutes -= 60;
			hours++;
		}
    while ( hours >= 24 ) {
			hours -= 24;
			days++;
		}
    while ( days >= 7 ) {
			days -= 7;
			weeks++;
		}
    if ( weeks > 0)
			cout << weeks << " week" << (weeks != 1 ? "s, " : ", ");
    if ( days > 0)
			cout << days << " day" << (days != 1 ? "s, " : ", ");
    if ( hours > 0)
			cout << hours << " hour" << (hours != 1 ? "s, " : ", ");
    if ( minutes > 0)
			cout << minutes << " minute" << (minutes != 1 ? "s, " : ", "); 
    cout << seconds << " second" << (seconds != 1 ? "s." : ".") << endl; 
  }
}
	
// The Lorenz system with solver----------------------------------------
class LorenzPoincareMap
{
public:
  typedef capd::poincare::PoincareMapJet<ITaylor> PoincareMap;
  
  IFunction section;
  IMap vectorField;
  ITaylor solver;
  PoincareMap pm;
  // Here we construct the Poincare map: The section is z=27, the
  // explicit times-step is 0.1, but ignored since dynamic step control 
  // is turned on by default. The 'PlusMinus' means that the section is 
  // crossed with 'z-27' changing sign from Plus to Minus. Other 
  // acceptable values are MinusPlus and None. The last means both 
  // directions are acceptable.
  LorenzPoincareMap (int order, interval _s, interval _r, interval _b)
    : section("var:x,y,z;fun:z-27;"),          
      vectorField("var:x,y,z;par:s,b,r;fun:s*(y-x),r*x-y-x*z,x*y-b*z;"), 
      solver(vectorField,order,interval(0.1)), 
      pm(solver,section,poincare::PlusMinus)                                                  
  {
    // Set parameter values s, r, and b
    vectorField.setParameter("s",_s);
    vectorField.setParameter("r",_r);
    vectorField.setParameter("b",_b);
  }
};

// This computes the Poincare map---------------------------------------
IVector extP(const IVector& u)
{
	// Define an instance of LorenzPoincareMap with classical parameters.
  int order  = ORDER;
  interval S = interval(10);
  interval R = interval(28);
  interval B = interval(8)/interval(3);
  LorenzPoincareMap local_pm(order,S,R,B);
  // u is assumed to be on the section
  // so it is 2-dim, represented by coordinates (x,y).
  // We simply add 27 as the third coordinate i.e. we embed the vector
  // into the full 3d-space
  IVector x(3);
  x[0] = u[0];
  x[1] = u[1];
  x[2] = R - interval(1);

  // We define a doubleton representation of the set
  C0Rect2Set s(x);
  // and compute one iteration
	x = local_pm.pm(s);
          
  // Project the 3d result into 2d:
  IVector result(2);
  result[0] = x[0];
  result[1] = x[1];
    
  return result;
 }
 
// This computes the derivative of the Poincare map---------------------
IMatrix extDP(const IVector& u)
{
	// Define an instance of LorenzPoincareMap with classical parameters.
  int order  = ORDER;
  interval S = interval(10);
  interval R = interval(28);
  interval B = interval(8)/interval(3);
  LorenzPoincareMap local_pm(order,S,R,B);

  // Again u is two dimensional, so embed it.
  IVector x(3);
  x[0] = u[0];
  x[1] = u[1];
  x[2] = R - interval(1);

  // For computing of derivative of PM we need a logarithmic norm
  IEuclLNorm N;
  // We define a doubleton representation of the set and its derivative
  // constructor sets initial condition for variational equation to Id.
  C1Rect2Set s(x,N);

  // Beware: the matrix monodromyMatrix will store derivatives of 
  // the FLOW, not of the Poincare map.
  IMatrix monodromyMatrix(3,3);
  x = local_pm.pm(s,monodromyMatrix);

  // This member function recomputes derivatives of the flow into 
  // derivatives of the Poincare map.
  IMatrix DP = local_pm.pm.computeDP(x,monodromyMatrix);

  // We extract from 3x3 matrix a 2x2 slice
  IMatrix result(2,2);
  result(1,1) = DP(1,1);
  result(1,2) = DP(1,2);
  result(2,1) = DP(2,1);
  result(2,2) = DP(2,2);
    
  return result;
}

// ---------------------------------------------------------------------
// The periodic shift  (x_1, x_2, ..., x_{2p}) -> (x_3, x_4, ..., x_2).
template <class T>
IVector shift(T v)
{
	int v_size = v.dimension();
	int period = v_size/2;
	IVector w(v_size);
	IVector u(2);
	u(1) = v(1);
	u(2) = v(2);
	for (int i = 1; i < period; ++i) {
		w(2*i-1) = v(2*i+1);
		w(2*i)   = v(2*i+2);
	}
	w(v_size-1) = u(1);
	w(v_size)   = u(2);
	return w;
}

// ---------------------------------------------------------------------
real vectorMax(RVector v)
{
	int v_size = v.dimension();
	real w = v(1);
	for (int i = 2; i <= v_size; ++i)
		w = max(w, v(i));
	return w;
}

// ---------------------------------------------------------------------
template <class T, class S>
S midpoint(T v)
{
	int v_size = v.dimension();
	S w(v_size);
	for (int i = 1; i <= v_size; ++i)
		w(i) = rightBound(v(i).mid());
	return w;
}

template <class T>
T mid(T v)
{
	int v_size = v.dimension();
	T w(v_size);
	for (int i = 1; i <= v_size; ++i)
		w(i) = v(i).mid();
	return w;
}

// ---------------------------------------------------------------------
RVector diameter(IVector v)
{
	int v_size = v.dimension();
	RVector w(v_size);
	for (int i = 1; i <= v_size; ++i)
		w(i) = rightBound(diam(v(i)));
	return w;
}

RMatrix diameter(IMatrix M)
{
	RMatrix N(M.numberOfRows(), M.numberOfColumns());
	for(int i=1;i<=M.numberOfRows();++i) {
    for(int j=1;j<=M.numberOfColumns();++j)
      N(i,j) = rightBound(diam(M(i,j)));
	}
	return N;
}

// ---------------------------------------------------------------------
bool verifyExistenceViaShooting(LorenzPoincareMap &pm, IVector gX0, IVector &NgX0)
{
	int period = gX0.dimension()/2;
	IVector gx0 = mid(gX0);
  IVector Pgx0(2*period);
  IVector xi0(2);
 	IVector Pxi0(2); 
	
  for (int i = 1; i <= period; ++i) {
		xi0(1) = gx0(2*i - 1);
		xi0(2) = gx0(2*i);
		
		// The Poincaré images.
		Pxi0 = extP(xi0);
		cout << "diam(Pxi0) = " << diameter(Pxi0) << endl;
		clock(SHOW_TIMING);
		Pgx0(2*i - 1) = Pxi0(1);
		Pgx0(2*i)     = Pxi0(2);
	}

	IVector sgX0 = shift(gx0);
	IVector    gF = sgX0 - Pgx0; // The global function.
	cout << "defect = " << midpoint<IVector, RVector>(gF) << endl;
#ifdef PRINTOUT
  cout << "--Done computing the defects: " << endl
			 << "gx0        = " << gx0 << endl
  		 << "Pgx0       = " << Pgx0 << endl
  		 << "s(gx0) - Pgx0 = " << gF << endl;
#endif 
 // At this point we have the approximative periodic orbit gx0 and its
 // enclosure gX0, together with the defect G.

  IMatrix DgF(2*period, 2*period);
	IVector X0(2);
	IMatrix DP(2,2);

  for (int j = 1; j <= period; ++j) {
		X0(1) = gX0(2*j - 1);
		X0(2) = gX0(2*j    );

		// The global function's derivatives.		
		DP = extDP(X0); 
		cout << "diam(DP) = " << diameter(DP) << endl;
		clock(SHOW_TIMING);
		DgF(2*j - 1, 2*j - 1) = -DP(1,1);
		DgF(2*j - 1, 2*j    ) = -DP(1,2);
		DgF(2*j    , 2*j - 1) = -DP(2,1);		
		DgF(2*j    , 2*j    ) = -DP(2,2);
		if ( j == period ) {
			DgF(2*j - 1, 1) = 1.0;
			DgF(2*j    , 2) = 1.0;
		}
		else {
			DgF(2*j - 1, 2*j + 1) = 1.0;
			DgF(2*j    , 2*j + 2) = 1.0;
		}
	}

#ifdef PRINTOUT
	cout << "DgF = " << endl << DgF << endl;
#endif
	
	// Apply the interval Newton operator
  NgX0 = gx0 - capd::matrixAlgorithms::gauss(DgF, gF);

  // Verification if N is a subset of X
#ifdef PRINTOUT  
  cout << "\n---------------------------------------------------\n\n" 
	     << "mid(N)   = " << midpoint(NgX0) << endl
	     << "mid(gX0) = " << midpoint(gX0) << endl;
  cout << "diam(N)   = " << diameter(NgX0) << endl;
  cout << "diam(gX0) = " << diameter(gX0) << endl;
#endif  
  if(subsetInterior(NgX0,gX0)) {
		cout << "N is a strict subset of gX0." << endl;
#ifdef PRINTOUT  
    cout << "the existence of period " << period 
			   << " orbit verified\n" << endl;
#endif
		return true;
	}
  else {
    cout << "N is not a subset of gX0. Intersecting..." << endl;
    IVector NcutX = intersection(NgX0, gX0);
#ifdef PRINTOUT
    cout << "intersection: " << NcutX << endl;
    cout << "N - gX0" << NgX0 - gX0 << endl;
#endif
		NgX0 = NcutX;
		return false;
  }
}

// ---------------------------------------------------------------------
IVector getData(istream& inFile, int period)
{
  IVector data(2*period);
  string oneRow;
  for (int i = 1; i <= period; i++)
    for (int j = 1; j <= 2; j++) {
      inFile >> oneRow;
      MpFloat a(oneRow,MpFloat::RoundNearest);
      data(2*(i-1) + j) = interval(a);
		}
  return data;
}

// ---------------------------------------------------------------------
string int2string(const int& number)
{
   ostringstream oss;
   oss << number;
   return oss.str();
}
 
// ---------------------------------------------------------------------
void readData(IVector &x0, int period, string fileName)
{
  //string  dataDir = "data/";
  
  // Construct the full file name.
  //string fullFileName = dataDir + fileName;
  string fullFileName = fileName;
  // Try to open the files.
  ifstream file(fullFileName.c_str(), ios_base::in);
  if ( !file ) {
    cout << "Cannot open file: '" << fullFileName << "' for reading" << endl;
    exit(1);
  }
  else {
    cout << "Opened " << fullFileName << " for reading...";
    x0 = getData(file, period);
    cout << " Done." << endl;
  }
}

// ---------------------------------------------------------------------
// Read all intermediate returns from file, and widen.
void getGlobalOrbitEnclosure(IVector &gX0, int period, 
														 double radius, string fileName)
{
	IVector gx0(2*period);
	readData(gx0, period, fileName);
  interval epsHull = radius*interval(-1,1);
	for (int i=1; i <= 2*period; ++i)
		gX0(i) = gx0(i) + epsHull;
}

//---------------------------- MAIN ------------------------------------

int main(int argc, char* argv[])
{
	if (argc != 5) {
		cout << "Syntax: ./wtrb radius precision period filename" << endl
				 << "   e.g. ./wtrb 1e-80 400 2 lor_2_BA.txt" << endl;
		exit(1);
	}
	cout << "Integrating with order " << ORDER << endl;

	double radius = atof(argv[1]);	// Set the size of the search area.
	int    BITS   = atoi(argv[2]);  // Set the multiprecision.
	int    period = atoi(argv[3]);  // Set the period of the orbit.
  string fileNm = argv[4];        // The filename with approximate data.
	double tolerance = 1e-100;      // Maximal stopping diameter.
	
	MpFloat::setDefaultPrecision(BITS);
	cout << "Computing with radius " << radius << endl;
	cout << "Computing with " << BITS << " bits precision." << endl;
  cout << "Searching for a period " << period << " orbit." << endl;

	try{
		// Define an instance of LorenzPoincareMap with classical parameters.
    int order = ORDER;
    interval s = interval(10);
    interval r = interval(28);
    interval b = interval(8)/interval(3);
    LorenzPoincareMap pm(order,s,r,b);

		IVector  gX0(2*period);
		IVector NgX0(2*period);
		getGlobalOrbitEnclosure(gX0, period, radius, fileNm);
		clock(START_TIMING);  
		bool keepOnComputing = true;
		while ( keepOnComputing) {
			keepOnComputing = verifyExistenceViaShooting(pm, gX0, NgX0);
			RVector diamNgX0 = diameter(NgX0);
			cout.precision(BITS/3);
			cout << "mid(N)     = " << midpoint<IVector, RVector>(NgX0) << endl;
			cout.precision(3);
			cout << "diam(N)    = " << diamNgX0 << endl;
			cout << "mid(N) - x < " << vectorMax(midpoint<IVector, RVector>(NgX0) - midpoint<IVector, RVector>(gX0)) << endl; 
			gX0 = NgX0;
			cout.precision(4);
			if ( vectorMax(diamNgX0) < tolerance ) {
				cout << "We have reached the desired tolerance." << endl;
				keepOnComputing = false;	
			}
			else {
				cout << "No convergence in first attempt! Quitting." << endl;
				return 0;
			}
		}
   }catch(exception& e)
  {
    cout << "\n\nException caught: "<< e.what() << endl;
  }
  clock(STOP_TIMING); 
  return 0;
} // END
